import json
import logging
import os
import pathlib
from dataclasses import dataclass
from typing import Dict, List, Optional

logger = logging.getLogger("magic_duckdb")

# Extra path is optional: you should have "dot" in your PATH. If not, you can set extra_path to the
# fully qualified path to your dot executable.
dot_path = None
# "c:\\Program files\\graphviz\\bin\\dot.exe"


def draw_graphviz(plan_json: str):
    try:
        # Defer loading, since this is optional
        import graphviz  # type: ignore # noqa

        logger.debug("Graphviz Python module is available")
    except Exception:
        logger.debug("Graphviz Python module not available")
        return None

    names_to_shapes = {
        "Query": "tripleoctagon",
        "RESULT_COLLECTOR": "doubleoctagon",
        "EXPLAIN_ANALYZE": "doubleoctagon",
        "PROJECTION": "rectangle",
        "DELIM_JOIN": "rectangle",
        "HASH_JOIN": "rectangle",
    }

    if dot_path is not None and os.path.exists(dot_path):
        logger.debug("Using dot_path: %s", dot_path)

        # graphviz.DOT_BINARY = pathlib.Path("c:\\Program files\\graphviz\\bin\\")
        # graphviz.backend.DOT_BINARY = pathlib.Path("c:\\Program files\\graphviz\\bin\\")
        graphviz.backend.dot_command.DOT_BINARY = pathlib.Path(dot_path)  # type: ignore # noqa

    dot = graphviz.Digraph()  # type: ignore # noqa

    @dataclass
    class Node:
        id: int
        name: str
        parent: "Node"
        properties: Dict

    nodes: List[Node] = []
    edges: List[Dict] = []

    def get_node(name, parent):
        id = len(nodes)
        node = Node(id=id, name=name, parent=parent, properties={})
        nodes.append(node)
        return node

    def process_node(node_json, parent: Optional[Node]):
        name = node_json.get("name")
        node = get_node(name, parent)

        node.properties["cardinality"] = node_json.get("cardinality")

        extra_info = node_json.get("extra_info")
        if extra_info is not None:
            node.properties["extra_info"] = str(extra_info).strip(" \t\r\n").replace(
                "\n", "\\n"
            )

        node.properties["result"] = node_json.get("result")

        timing = node_json.get("timing")
        timing_percent = (
            timing / total_time if timing is not None and total_time is not None else 0
        )

        if timing is not None:
            node.properties["timing"] = f"{timing:.2f} ({timing_percent:.0%})"
        props = [
            f"{k}={v}" if k != "extra_info" else v
            for k, v in node.properties.items()
            if v is not None and v != ""
        ]
        propstr = "\\n".join(props)

        shape = names_to_shapes.get(node.name)
        if shape is None:
            shape = "ellipse"
        dot.node(f"{node.id}", f"{node.name}\\n{propstr}", weight="10", shape=shape)

        if parent is not None:
            edges.append((parent.id, node.id, {"weight": node_json.get("timing")}))  # type: ignore

            dot.edge(f"{parent.id}", f"{node.id}")

        children = node_json.get("children")
        if children is not None:
            for child in children:
                process_node(child, node)

    j = json.loads(plan_json)

    total_time = j.get("timing")
    process_node(j, None)

    return dot



#
#
# Futures:
# Use embeddings
# Maintain conversation context
import logging
import textwrap
from typing import Optional, Tuple

from openai import OpenAI  #

logger = logging.getLogger("magic_duckdb")

OPENAI_KEY = None
print_prompts = False

__LAST_RESULT = None
COMPLETION_ENGINE = "gpt-3.5-turbo"
# COMPLETION_ENGINE = "gpt-4-32k-0314"

# CHATCOMPLETION_ENGINE = "gpt-4-0125-preview"

__OPENAI_CLIENT = None


def get_client():
    if OPENAI_KEY is None:
        raise ValueError(
            "Set the OPENAI_KEY before using. \nfrom magic_duckdb.extras import sql_ai\nsql_ai.OPENAI_KEY=yourkey"
        )

    global __OPENAI_CLIENT
    if __OPENAI_CLIENT is None:
        __OPENAI_CLIENT = OpenAI(api_key=OPENAI_KEY)

    return __OPENAI_CLIENT


def get_columns(connection) -> str:
    df = connection.sql(
        "select table_name, column_name, data_type from duckdb_columns"
    ).df()

    return df.to_csv(index=False)
    col_desc = []

    return ["column_name", "data_type"]

    for t in df["table_name"].unique():
        cols = [
            f"{v[0]} (type = {v[1]})"
            for v in df.loc[df["table_name"] == t, ["column_name", "data_type"]].values
        ]
        cols_desc = ",".join(cols)

        desc = f"Table {t} has the following columns and data types - {cols_desc}"
        col_desc.append(desc)

    return "\n".join(col_desc)


def get_schema(connection) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    try:
        if connection is None:
            return None, None, None

        t = connection.sql("PRAGMA show_tables")
        if t is None:  # no tables = None
            return None, None, None
        else:
            tables = t.df()

        if len(tables) == 0:
            tables = None
        else:
            tables = ", ".join(tables["name"].values)

        cols = get_columns(connection)

        constraints = connection.sql("select * from duckdb_constraints").df()

        if len(constraints) == 0:
            constraints = None
        else:
            constraints = constraints.to_csv()

        return tables, cols, constraints
    except Exception:
        logger.exception("Error getting schema")
        return None, None, None


def call_ai(connection, chat: bool, prompt: str, query: str):
    return ai_statement(connection=connection, prompt=prompt, statement=query)


def exec_ai_prompt(prompt: str, engine=None) -> str | None:
    client = get_client()

    completion = client.chat.completions.create(
        model=COMPLETION_ENGINE if engine is None else engine,
        messages=[
            {
                "role": "user",
                "content": prompt,
            }
        ],
    )
    message = completion.choices[
        0
    ].message  # response["choices"][0]["message"]["content"]  # type: ignore

    global __LAST_RESULT
    __LAST_RESULT = message
    result: str | None = message.content
    return result


def ai_statement(connection, prompt: str, statement: str):
    logger.info("Passing %s statement to AI ): %s", prompt, statement)
    # Prepare prompt
    tables, cols, constraints = get_schema(connection)

    prompt = f"{prompt}\nMy query is: {statement}\nMy database is DuckDB. DuckDB's SQL is similar to postgresql. DuckDB sql supports: select, from, join, where, group by, grouping sets, having, order by, limit, sample, unnest, with, window, qualify, values and filter. "
    context = f"I am writing SQL for a DuckDB database. My database's tables, columns and column data types are the following comma separated table: \n{cols}\n\nConstraints: {constraints}"

    full_prompt = context + "\nMy question is: " + prompt
    logger.debug("Num tokens: %s", len(prompt.split(" ")))

    logger.info("Prompt = \n %s", full_prompt)
    if print_prompts:
        print("-------------Prompt---------------")  # noqa: T201
        print(full_prompt)  # noqa: T201

    result = exec_ai_prompt(full_prompt)

    cell = textwrap.dedent(result) if result is not None else ""
    # Insert 4 spaces of indentation before each line
    cell = textwrap.indent(cell, " " * 4)

    print("-------------OpenAI Response---------------")  # noqa: T201
    print(cell)  # noqa: T201
    return cell



import logging
from typing import List

from pandas import DataFrame

from magic_duckdb import magic

logger = logging.getLogger("magic_duckdb")

sql_expects_tablename = [
    "UNION",
    "UNION ALL",
    "UNION ALL BY NAME",
    "UNION BY NAME",
    "JOIN",
    "INNER JOIN",
    "LEFT JOIN",
    "RIGHT JOIN",
    "FULL JOIN",
    "LEFT OUTER JOIN",
    "RIGHT OUTER JOIN",
    "FROM",
    "INTO",
]

sql_phrases = [
    "PRAGMA",
    "SELECT",
    "WHERE",
    "GROUP BY",
    "ORDER BY",
    "LIMIT",
    "INSERT",
    "UPDATE",
    "DELETE",
    "ALTER",
    "DROP",
    "TRUNCATE",
    "TABLE",
    "DATABASE",
    "INDEX",
    "VIEW",
    "FUNCTION",
    "PROCEDURE",
    "TRIGGER",
    "AND",
    "OR",
    "NOT",
    "BETWEEN",
    "LIKE",
    "IN",
    "NULL",
    "IS",
    "EXISTS",
    "COUNT",
    "SUM",
    "MIN",
    "MAX",
    "AVG",
    "DISTINCT",
    "AS",
    "CREATE TABLE",
    "CREATE OR REPLACE TABLE",
    "CREATE TABLE IF NOT EXISTS",
    "CREATE VIEW",
]

pragma_phrases = [
    "PRAGMA version",
    "PRAGMA database_list",
    "PRAGMA database_size",
    "PRAGMA show_tables",
    "PRAGMA show_tables_expanded",
    "PRAGMA table_info('",
    "PRAGMA functions",
    "PRAGMA collations",
    "PRAGMA enable_progress_bar",
    "PRAGMA disable_progress_bar",
    "PRAGMA enable_profiling",
    "PRAGMA disable_profiling",
    "PRAGMA disable_optimizer",
    "PRAGMA enable_optimizer",
    "PRAGMA enable_verification",
    "PRAGMA disable_verification",
    "PRAGMA verify_parallelism",
    "PRAGMA disable_verify_parallelism",
    "PRAGMA force_index_join",
    "PRAGMA force_checkpoint",
]


def get_table_names(ipython) -> List[str]:
    try:
        user_keys = [k for k, v in ipython.user_ns.items() if isinstance(v, DataFrame)]

        if magic.connection is not None:
            tables = magic.connection.sql("show tables")
            if tables is None:
                return user_keys
            else:
                return list(tables.df()["name"]) + user_keys
        else:
            logger.info(user_keys)
            return user_keys
    except Exception:
        logger.exception("Unable to get table names")
        return []


def get_column_names(ipython, tablename: str) -> List[str]:
    try:
        # check if an object
        # o = ipython.ev(tablename)
        o = ipython.user_ns.get(tablename)
        if o is None and magic.connection is not None:
            columns = magic.connection.sql(f"pragma table_info('{tablename}')")
            if columns is None:
                logger.debug("None columns")
                return []
            else:
                names = list(columns.df().astype(str)["name"])
                logger.debug("Column names: %s", names)
                return names
        elif o is not None:
            if isinstance(o, DataFrame):
                return list(o.columns)
            else:
                logger.debug(
                    "%s in namespace, but not a DataFrame %s", tablename, type(o)
                )
                return []
        else:
            return []
    except Exception:
        logger.exception("Unable to get column names")
        return []


def init_completers(ip):
    try:
        from magic_duckdb.autocomplete.autocompletion_v2 import init_completer

        init_completer(ipython=ip)
    except Exception as e:
        logger.debug(
            "Unable to initialize autocompletion_v2. iPython 8.6.0+ is required for autocomplete. %s", e
        )



import logging
import pathlib
from subprocess import PIPE, STDOUT, Popen

logger = logging.getLogger("magic_duckdb")

language_mode = "postgresql"
indentStyle = "standard"  # noqa: N816
keywordCase = "upper"  # noqa: N816
linesBetweenQueries = 0  # noqa: N816
logicalOperatorNewline = "after"  # noqa: N816
denseOperators = "true"  # noqa: N816


def formatsql(query: str):
    logger.debug("Formatting: %s", query)
    config = f'"keywordCase": "{keywordCase}", "indentStyle": "{indentStyle}", "linesBetweenQuerys": {linesBetweenQueries}, "logicalOperatorNewline": "{logicalOperatorNewline}", "denseOperators": "{denseOperators}"'
    config = "{" + config + "}"

    pathlib.Path("format_config.json").write_text(config)

    cmd = [
        "npx",
        "sql-formatter",
        "-l",
        language_mode,
        "--config",
        "format_config.json",
    ]

    p = Popen(cmd, stdout=PIPE, stdin=PIPE, stderr=STDOUT, shell=True)  # noqa: S602
    stdout_data = p.communicate(input=query.encode())[0]

    print(stdout_data.decode())  # noqa: T201
    return stdout_data.decode()



import argparse
import logging
from pathlib import Path
from typing import Dict, Optional

# from IPython.core.getipython import get_ipython
from duckdb import ConnectionException, DuckDBPyConnection
from IPython.core.magic import (
    Magics,
    cell_magic,
    line_magic,
    magics_class,
    needs_local_scope,
    no_var_expand,
)
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from traitlets.config.configurable import Configurable

from . import MAGIC_NAME
from .duckdb_mode import DuckDbMode
from .extras import jinja_template

logger = logging.getLogger("magic_duckdb")

# dbwrapper: To override database logic, replace or monkeypatch this object
dbwrapper: DuckDbMode = DuckDbMode()


def _get_obj_from_name(shell, name: str) -> Optional[object]:
    return shell.ev(name) if shell is not None else None


@magics_class
class DuckDbMagic(Magics, Configurable):
    # selected via -t. None = Pandas.
    default_export_function = None
    # database connection object created via -d (default), -cn (connection string) or -co (connection object)
    connection: Optional[DuckDBPyConnection] = None

    def __init__(self, shell, attr_matches: bool = True):
        Configurable.__init__(self, config=shell.config)
        Magics.__init__(self, shell=shell)

        # Add to configurable modules via %config
        self.shell.configurables.append(self)  # type: ignore

    def connect_by_objectname(self, connection_object):
        con: DuckDBPyConnection = _get_obj_from_name(self.shell, connection_object)  # type: ignore
        if not isinstance(con, DuckDBPyConnection):
            raise ValueError(f"{connection_object} is not a DuckDBPyConnection")
        elif con is None:
            raise ValueError(f"Couldn't find {connection_object}")
        else:
            logger.info("Using existing connection: %s", connection_object)

            self.connection = con

    def format_wrapper(self, query: str):
        try:
            from magic_duckdb.extras.sqlformatter import formatsql

            return formatsql(query)
        except Exception as e:
            logger.exception("Error processing")
            raise e

    def ai_wrapper(self, chat: bool, prompt: str, query: str):
        try:
            from magic_duckdb.extras import sql_ai

            sql_ai.call_ai(self.connection, chat, prompt, query)
            return None  # suppress for now, need to figure out formatting
        except Exception as e:
            logger.exception("Error with AI")
            raise e

    @no_var_expand
    @needs_local_scope
    @line_magic(MAGIC_NAME)
    @cell_magic(MAGIC_NAME)
    @magic_arguments()
    @argument("-l", "--listtypes", help="List the available types", action="store_true")
    @argument("-g", "--getcon", help="Return current connection", action="store_true")
    @argument(
        "-d", "--default_connection", help="Use default connection", action="store_true"
    )
    @argument("-f", "--format", help="Format (beautify) SQL input", action="store_true")
    @argument("-ai", help="OpenAI", action="store_true")
    @argument("-aichat", help="OpenAI Chat", action="store_true")
    @argument(
        "-cn",
        "--connection_string",
        help="Connect to a database using the connection string, such as :memory: or file.db",
        nargs=1,
        type=str,
        action="store",
    )
    @argument(
        "-r",
        "--readfile",
        help="Read SQL from a file",
        nargs=1,
        type=str,
        action="store",
    )
    @argument(
        "-co",
        "--connection_object",
        help="Connect to a database using the connection object",
        nargs=1,
        type=str,
        action="store",
    )
    @argument(
        "-t",
        "--type",
        help="Set the default output type",
        nargs=1,
        type=str,
        action="store",
    )
    @argument(
        "-o",
        "--output",
        help="Write the output to the specified variable",
        nargs=1,
        type=str,
        action="store",
    )
    @argument(
        "-e",
        "--explain",
        help="Explain or Explain Analyze or AST",
        nargs=1,
        type=str,
        action="store",
    )
    @argument("--tables", help="Return table names", action="store_true")
    @argument("--close", help="Close database connection", action="store_true")
    @argument("-j", "--jinja2", help="Apply Jinja2 Template", action="store_true")
    @argument(
        "-p",
        "--param",
        dest="queryparams",
        help="Params from user_ns",
        action="append",
    )
    @argument(
        "-tp",
        "--type_param",
        dest="type_param",
        nargs=2,
        help="Params to pass to the Output Type function, given as a name/value pair",
        action="append",
    )
    @argument("rest", nargs=argparse.REMAINDER)
    def execute(self, line: str = "", cell: str = "", local_ns=None):
        cell = "" if cell is None else cell
        line = "" if line is None else line
        user_ns: Dict[str, object] = self.shell.user_ns  # type: ignore

        args = parse_argstring(self.execute, line)

        rest = " ".join(args.rest)

        if args.readfile:
            sql_file_path = Path(args.readfile[0])
            query = sql_file_path.read_text()
        else:
            query = f"{rest}\n{cell}".strip()

        if args.jinja2:
            query = jinja_template.apply_template(query, user_ns)

        if args.listtypes:
            return dbwrapper.export_functions
        elif args.getcon:
            return self.connection
        elif args.format:
            return self.format_wrapper(query)
        elif args.ai:
            return self.ai_wrapper(False, rest, query)
        elif args.aichat:
            return self.ai_wrapper(False, rest, query)
        elif args.close:
            if self.connection is not None:
                try:
                    self.connection.close()
                finally:
                    self.connection = None

        export_kwargs = {}

        if args.type_param:
            for n, v in args.type_param:
                if str(v).isdigit():
                    v = int(v)
                export_kwargs[n] = v

        thisconnection = None
        if args.default_connection:
            thisconnection = dbwrapper.default_connection()
        if args.connection_object:
            thisconnection = self.connect_by_objectname(args.connection_object[0])
        if args.connection_string:
            thisconnection = dbwrapper.connect(args.connection_string[0])

        if args.type and args.type[0] not in dbwrapper.export_functions:
            raise ValueError(
                f"{args.type[0]} not found in {dbwrapper.export_functions}"
            )

        if args.explain:
            explain_function = args.explain[0]
            if explain_function not in dbwrapper.explain_functions:
                raise ValueError(
                    f"{explain_function} not found in {dbwrapper.explain_functions}"
                )
        else:
            explain_function = None

        logger.debug("Query = %s", query)
        if query is None or len(query) == 0:
            if args.type:
                # print("Default format changed: {args.type[0]}")
                self.default_export_function = args.type[0]
            if thisconnection is not None:
                # print(f"Default connection changed: ", "default_connection()" if args.default_connection else args.connection_object[0] if args.connection_object else args.connection_string)
                self.connection = thisconnection
            logger.debug("Nothing to execute")
            return

        if thisconnection is None:
            if self.connection is None:
                # print("Default connection changed: default_connection()")
                self.connection = dbwrapper.default_connection()
            thisconnection = self.connection

        try:
            if args.queryparams:
                queryparams = [user_ns[p] for p in args.queryparams]
            else:
                queryparams = None

            if args.tables:
                return thisconnection.get_table_names(query)  # type: ignore

            o = dbwrapper.execute(
                query_string=query,
                connection=thisconnection,
                export_function=(
                    args.type[0] if args.type else self.default_export_function
                ),
                explain_function=explain_function,
                params=queryparams,
                export_kwargs=export_kwargs,
            )
            if args.output:
                user_ns[args.output[0]] = o

            return o

        except ConnectionException:
            logger.exception(
                "Unable to connect, connection may already be closed. Setting connection to None"
            )
            self.connection = None
